﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.ProgramSynthesis.Push.Generators {
  using HeuristicLab.Core;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Configuration;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Data.Pool;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Expressions;
  using HeuristicLab.Random;

  public class LinearCodeGenerator {
    public static PushProgram RandomProgram(IManagedPool<PushProgram> pool, IManagedPool<PooledList<Expression>> expressionListPool, int maxPoints, IRandom random = null, IReadOnlyPushConfiguration pushGpConfiguration = null, IDictionary<string, Expression> customExpressions = null) {
      var code = RandomCode(maxPoints, expressionListPool, random, pushGpConfiguration, customExpressions);

      return PushProgram.Create(pool, code);
    }

    public static PushProgram RandomProgram(IManagedPool<PooledList<Expression>> expressionListPool, int maxPoints, IRandom random = null, IReadOnlyPushConfiguration pushGpConfiguration = null, IDictionary<string, Expression> customExpressions = null) {
      var code = RandomCode(maxPoints, expressionListPool, random, pushGpConfiguration, customExpressions);

      return new PushProgram(code);
    }

    public static PushProgram RandomProgram(int maxPoints, IRandom random = null, IReadOnlyPushConfiguration pushGpConfiguration = null, IDictionary<string, Expression> customExpressions = null) {
      var code = RandomCode(maxPoints, null, random, pushGpConfiguration, customExpressions);

      return new PushProgram(code);
    }

    public static IReadOnlyList<Expression> RandomCode(
      int maxPoints,
      IManagedPool<PooledList<Expression>> expressionListPool = null,
      IRandom random = null,
      IReadOnlyPushConfiguration pushGpConfiguration = null,
      IDictionary<string, Expression> customExpressions = null) {
      if (maxPoints == 0)
        return new Expression[0];

      random = random ?? new FastRandom();
      pushGpConfiguration = pushGpConfiguration ?? new PushConfiguration();

      var size = maxPoints <= 1 ? 1 : random.Next(1, maxPoints);
      var expressions = expressionListPool == null ? new List<Expression>(size) : expressionListPool.Get();

      for (var i = 0; i < size; i++) {
        var expression = random.NextDouble() <= pushGpConfiguration.ErcProbability
         ? CreateRandomErcExpression(random, pushGpConfiguration)
         : CreateExpression(random, pushGpConfiguration, customExpressions);

        expressions.Add(expression);
      }

      return expressions;
    }

    private static Expression CreateExpression(IRandom random, IReadOnlyPushConfiguration pushGpConfiguration, IDictionary<string, Expression> customExpressions = null) {
      var customCount = customExpressions == null ? 0 : customExpressions.Count - 1;
      var index = random.Next(0, pushGpConfiguration.EnabledExpressions.Count + customCount - 1);

      if ((index >= 0) && (index < pushGpConfiguration.EnabledExpressions.Count)) {
        var name = pushGpConfiguration.EnabledExpressions[index];
        return ExpressionTable.GetExpression(name);
      }

      if (customExpressions == null)
        throw new IndexOutOfRangeException();

      return customExpressions.ElementAt(index - (pushGpConfiguration.EnabledExpressions.Count - 1)).Value;
    }

    private static Expression CreateRandomErcExpression(IRandom random, IReadOnlyPushConfiguration pushGpConfiguration) {
      var value = random.Next(0, 3);

      switch (value) {
        case 0:
          return new IntegerPushExpression(random.Next(
            pushGpConfiguration.MinRandomInteger,
            pushGpConfiguration.MaxRandomInteger));

        case 1:
          return new FloatPushExpression(
            random.NextDouble() *
            (pushGpConfiguration.MaxRandomFloat - pushGpConfiguration.MinRandomFloat) +
            pushGpConfiguration.MinRandomFloat);

        case 2:
          return new BooleanPushExpression(random.NextDouble() > 0.5);

        case 3:
          return new NamePushExpression(NameGenerator.RandomName(random));

        default:
          throw new NotImplementedException();
      }
    }
  }
}
